﻿//================================================================================
//  
//       FreqCounter.cs      周波数カウンター
//  
//        (c)2012/08/28  by OjisanKoubou
//  
//================================================================================
//  
//  このプログラムは Ojisankoubou( OGAWA Ichiro ) が著作権を持っています。
//  個人の研究目的には、修正、改造など自由に使用できます。
//  発表される際は著作権表示部分を消さないようにしてください。
//  このソフトウェアの全部もしくは一部を製品に組み込んだり販売したりするなどの
//  商業的利用はできません。
//

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;

namespace APB_3
{
    //
    //WRITE
    //  0x0800   7 6 5 4 3 2 1 0
    //                         +-- START/STOP  0 を書くと測定停止、1 を書くと測定開始　
    //
    //  0x0801   7 6 5 4 3 2 1 0
    //             | | | | +-+-+-- MESBITSEL   0:1/1  1:1/2    2:1/4   ...  7:1/128　
    //             | | +-+-------- PRESCALER   0:1/1  1:1/256  2:1/65536
    //             | +------------ FCINSEL     周波数カウンター入力切替  0:直接 1:周波数変換後
    //             +-------------- SEQ         0:単発測定   1: 連続測定
    //
    //  0x0802   7 6 5 4 3 2 1 0
    //             | | | +-+-+-+-- SIGSEL     入力切替　
    //             +-+-+---------- FMCICSEL   FM 検波器 CIC 段数切替  0:128 1:256 ... 5:4096
    //
    //
    //READ
    //  0x0800   7 6 5 4 3 2 1 0
    //                     | +-+-- STATE 0:測定終了 1:測定開始 2:開始カウンター値取得 or 連続測定中
    //                     +------ FSIGN(2) STATE(1..0) FSIGN:周波数の符号 0:正 1:負
    //
    //  0x0801   SEQCNT   (  7 downto  0 )
    //  0x0802   MESCNT   (  7 downto  0 )
    //  0x0803   CNT_STOP (  7 downto  0 )
    //  0x0804   CNT_STOP ( 15 downto  8 )
    //  0x0805   CNT_STOP ( 23 downto 16 )
    //  0x0806   CNT_STOP ( 31 downto 24 )
    //  0x0807   CNT_START(  7 downto  0 )
    //  0x0808   CNT_START( 15 downto  8 )
    //  0x0809   CNT_START( 23 downto 16 )
    //  0x080A   CNT_START( 31 downto 24 )
    //
    
    class FreqCounter
    {
        public static int mescnt = 0;
        public static int  seqcnt = 0;
        public static long cnt = 0;
        
        
        //
        // 測定開始
        //
        // 実行前に APB3 の使用権を得ること
        public static void Start(int source, int mescnt_, int seq)
        {
            mescnt = mescnt_;
            int setting = (source << 8) + (seq << 6) + mescnt;
            Apb3.Fpga.FpgaWrite(Apb3.Fpga.freqdet+1, 2, setting);  // 周波数カウンター設定
            Apb3.Fpga.FpgaWrite(Apb3.Fpga.freqdet, 1);          // 周波数測定開始
            if( seq==1 ) {
                Byte[] Rbuf = new Byte[256];
                int timeout = 100;
                while( true ) {
                    Apb3.Fpga.FpgaRead(Apb3.Fpga.freqdet, 11, Rbuf);
                    if ((Rbuf[0] & 0x03) == 2) break;   //cnt 値設定検出
                    Thread.Sleep(100);
                    if (timeout-- == 0) return;
                }
                seqcnt = Rbuf[1];;
                cnt = 0;
                for (int i = 10; i > 6; i--) cnt = (cnt << 8) + Rbuf[i];
            }
        }
        
        public static void Start(int source, int mescnt)
        {
            Start(source, mescnt, 0);
        }
        
        public static void Start(int source)
        {
            Start(source, 0, 0);
        }
        
        
        //
        // 測定終了
        //
        // 実行前に APB3 の使用権を得ること
        public static void Stop()
        {
            Apb3.Fpga.FpgaWrite(Apb3.Fpga.freqdet, 0);          // 周波数測定終了
        }
        
        
        //
        // Time 時間で測定できる mescnt 値を計算する
        //
        // 実行前に APB3 の使用権を得ること
        public static int Mescnt( int source, double time )
        {
            Start( source );            // 分周なしで予備周波数測定  
            double freq = Measure();
            if (freq < 0 ) return (int)freq;
            int mescnt = (int)(Math.Log10(time * freq) / 0.301);
            if (mescnt > 23) mescnt = 23;
            if (mescnt < 0 ) mescnt = 0;
            return mescnt;
        }
        

        //
        // 周波数測定実行（非測定信号を 2^mescnt 分周した１周期測って周波数測定）
        //
        // 実行前に APB3 の使用権を得ること
        public static double Measure()
        {
            double freq = -1;
            Byte[] Rbuf = new Byte[256];
            int timeout = 40;       // 2秒でタイムアウト
            while (true)            // 測定終了待ち
            {
                if( Apb3.Fpga.FpgaRead(Apb3.Fpga.freqdet, 11, Rbuf) < 0 ) return error.usb_disconnect;
                if ((Rbuf[0] & 0x03) == 0) break;   //測定終了検出
                Thread.Sleep(50);
                if (timeout-- == 0) {
                    Apb3.mainf.LogPrint("\nFreqCounter.Measure: timeout");
                    return error.timeout;
                }
            }
            
            long stop = 0;
            for (int i = 6; i > 2; i--) stop = (stop << 8) + Rbuf[i];
            long start = 0;
            for (int i = 10; i > 6; i--) start = (start << 8) + Rbuf[i];
            
            long diff = stop - start;
            if (diff < 0) diff += 0x100000000;
            freq = Apb3.clockfreq / diff * Math.Pow(2, mescnt);
            
            Apb3.mainf.LogPrint("\nstop=" + stop.ToString() + "  start=" + start.ToString());
            Apb3.mainf.LogPrint("\nmescnt=" + mescnt.ToString() + "  diff=" + diff.ToString() + "  freq=" + freq.ToString());
            return freq;
        }
        
        
        //
        // 連続周波数測定     update が true だと seqcnt, cnt を更新する
        //
        // 実行前に APB3s の使用権を得ること
        public static double SeqMeasure( bool update )
        {
            int timeout = 40;              // 2秒でタイムアウト
            while( true ) {
                Byte[] Rbuf = new Byte[32];
                if( Apb3.Fpga.FpgaRead(Apb3.Fpga.freqdet, 32, Rbuf) < 0 ) return error.usb_disconnect;
                
                if ((Rbuf[0] & 0x03) == 0x02)    //連続測定中か？
                {
                    int seqcnt1 = Rbuf[1];
                    int seqcnt2 = Rbuf[16+1];
                    long cnt1= 0, cnt2=0;
                    for (int i = 6; i > 2; i--) cnt1 = (cnt1 << 8) + Rbuf[i];
                    for (int i = 22; i > 18; i--) cnt2 = (cnt2 << 8) + Rbuf[i];
                    
                    
                    if( ( seqcnt1 != seqcnt )&&( seqcnt1 == seqcnt2 )&&( cnt1 == cnt2 ) )
                    {   // seqcnt が変化し、さらに2回読み込みしたデータが一致していたら正常に測定できた
                        long diff = cnt1 - cnt;
                        if (diff < 0) diff += 0x100000000;
                        
                        int seqc = seqcnt1-seqcnt;
                        if( seqc < 0 ) seqc += 256;
                        
                        double freq = Apb3.clockfreq / diff * Math.Pow(2, mescnt) * seqc;
                        
Apb3.mainf.LogPrint("\nfreq=" + freq.ToString() + "  cnt1=" + cnt1.ToString() + "  cnt=" + cnt.ToString() + "  seqcnt1=" + seqcnt1.ToString() + "  seqcnt=" + seqcnt.ToString() );
                        
                        if( update ) {
                            seqcnt = seqcnt1;
                            cnt = cnt1;
                        }
                        return freq;
                    }
                }
                if (timeout-- == 0) return error.timeout;
                Thread.Sleep(50);
            }
        }
        
        
    }
}
